#!/usr/bin/perl
use Carp qw(confess);
use Data::Dumper;
use Getopt::Long;
use PDF::API2::Lite;
use strict;
use warnings;

sub Debug                ( @     );
sub DrawBorders          ( $$$$$ );
sub DrawFill             ( $$    );
sub DrawLine             ( $$$$$ );
sub DrawRegion           ( $$    );
sub DrawText             ( $$$   );
sub FillColor            ( $     );
sub Lookup               ( $     );
sub ParseCharacterData   ( $     );
sub ParseFormDefinition  ( $     );
sub ParseTextData        ( $     );
sub Print                (       );
sub PrintCharacterData   (       );
sub PrintFormTranslation (       );
sub StrokeColor          ( $     );

our $File;			# File name being processed.
our $Form;			# The form file being processed.
our $ID          = 0;		# The Form ID for duplicate pages.
our $P;                         # The PDF object.
our $Pass        = "";		# The render pass we are currently on.
our %CharData    = ();          # $CharData{$Form.$ID:$field}
our %Config      = ();		# Internal constants.
our %Fonts       = ();		# The font storage for loaded fonts.
our %Opt         = ();		# Command-line options or defaults.
our %Region      = ();          # $Region{$Form:$region}
our %Text        = ();          # $Text{$Form:$key}
our @LookupStack = ();          # Used to avoid recursive lookups.
our @Pages       = ();		# ("$Form.$ID:$region", ...)
our @RenderStack = ();          # Used to avoid recursive renders.

$Opt{Baseline}  = 0.03*72;
$Opt{DataDir}   = "./";
$Opt{Debug}     = 0;
$Opt{Error}     = 0;
$Opt{FontAlign} = "center";
$Opt{FontDir}   = "./";
$Opt{FontSize}  = 9;
$Opt{FontType}  = "DaxCondensed-Medium";
$Opt{LineWidth} = 0.005*72;
$Opt{Padding}   = 0.03*72;
$Opt{Paper}     = "Letter";
$Opt{Printer}   = "Optra";
$Opt{TmpDir}    = "/tmp";

$Config{Paper}{Letter}= {
			 width  =>  8.5*72,
			 height => 11.0*72,
			 margin =>  0.5*72
			};
$Config{Keys}{Opts} = {
		       r  => 'row',
		       b  => 'border',
		       bl => 'border_left',
		       bt => 'border_top',
		       br => 'border_right',
		       bb => 'border_bottom',
		       cb => 'cell_border',
		       cbb=> 'cell_border_bottom',
		       cbt=> 'cell_border_top',
		       cbl=> 'cell_border_left',
		       cbr=> 'cell_border_right',
		       a  => 'text_align',
		       s  => 'text_size',
		       f  => 'text_font',
		       c  => 'col',
		       w  => 'width',
		       h  => 'height',
		       fg => 'foreground',
		       bg => 'background',
		       i  => 'reverse_color'
		      };
$Config{MissingRegion} = {
			  text_size     => 5,
			  color         => "grey50",
			  reverse_color => 1,
			  text_align    => 'left'
			 };
#=============================================================================
#=============================================================================
# Main ()

GetOptions (\%Opt, "Debug");

$Data::Dumper::Indent = 1;

$SIG{INT}  = sub { confess "Interrupted"; };

FILE: for $File (@ARGV) {
  ParseCharacterData $File
    or warn "Error parsing $File" and next FILE;
  #print Dumper \%Text;
  Print;
}
exit 0;

#=============================================================================
#=============================================================================
# General utility.
sub PageWidth  (    ) { return $Config{Paper}{$Opt{Paper}}{width};  }
sub PageLength (    ) { return $Config{Paper}{$Opt{Paper}}{height}; }
sub PageUnits  (    ) { return $Config{Paper}{$Opt{Paper}}{unit};   }
sub Margin     (    ) { return $Config{Paper}{$Opt{Paper}}{margin}; }
sub Point      ( $$ ) { return $_[0], PageLength-$_[1];             }
sub Date       ( ;$ ) {
  my @time = localtime (shift || time);
  return sprintf "%4d.%02d.%02d", $time[5]+1900 ,$time[4]+1,$time[3];
}

#=============================================================================
#=============================================================================
# Parse the various files.

#-----------------------------------------------------------------------------
# Read the character's data.
sub ParseCharacterData ( $ ) {
  my $file = shift;
  open CHAR, $file or die;
 LINE: while (<CHAR>) {
    chomp;
    s/\s+$//;

    # Character file with optional $ID.
    Debug "Line:", $_;
    if ( /^\#\s*([-\w]+)\.(\w+)$/ ) {
      $Form = $1;
      my $Language = $2;
      $ID++;
      Debug "New file, Form:", $Form, "Lang:", $Language, "ID:", $ID;
      # Only parse the form the first time.
      if (not grep /^$Form\./, @Pages) {
	ParseFormDefinition $Form or die;
	ParseTextData "$Form.$Language" or die;
      }
      push @Pages,
	map {s/^$Form:/$Form.$ID:/; $_}
	  grep /^$Form:page(\d+)$/,
	    keys %Region;
      next LINE;
    }

    /^\s*(\w+)\s*=\s*(.*?)\s*$/ or next;
    die "Values in character before form defined.\n"
      if not defined $Form or not defined $ID;
    my ($key, $value) = (lc $1, $2);
    Debug "Form:", $Form, "ID:", $ID, "key:", $key;
    $key = "$Form.$ID:$key";
    $CharData{$key} = [] if not exists $CharData{$key};
    push @{ $CharData{$key} }, $value;
  }

  close CHAR;
  return 1;
}

#-----------------------------------------------------------------------------
# Read the form itself.
sub ParseFormDefinition ( $ ) {
  my $form = shift;
  my $file = "$Opt{DataDir}/$form";
  open DEF, $file or die;
  my $region;
  my ($last_row,$last_col);
 LINE: while (<DEF>) {
    chomp;
    s/^\#.*//;
    s/\s+$//;
    next unless length $_;
    if (/^(\w+):\s*(.*?)\s*$/) {
      # Region start.
      $region = "$Form:$1";
      $Region{$region}{Option}{foreground} = 'black';
      $Region{$region}{Option}{background} = '';
      for my $opt (split /\s+/, $2) {
	$opt =~ /^(\w+)=(.+)$/ or warn "Unparsable $opt" and next;
	my ($key, $value) = ($1, $2);
	$key = $Config{Keys}{Opts}{$key} if exists $Config{Keys}{Opts}{$key};
	$Region{$region}{Option}{$key} = $value;
      }
      $last_row = 1;
      $last_col = 1;
    } elsif (/^\s+/) {
      s/^\s+//;
      my @opts = split;
      my $field = shift @opts;
      my $ref = {
		 height        => 1,
		 width         => 1,
		 row           => $last_row,
		 col           => $last_col,
		 content       => $field,
		 border        => $Region{$region}{Option}{cell_border},
		 border_bottom => $Region{$region}{Option}{cell_border_bottom},
		 border_top    => $Region{$region}{Option}{cell_border_top},
		 border_left   => $Region{$region}{Option}{cell_border_left},
		 border_right  => $Region{$region}{Option}{cell_border_right},
		 text_align    => $Region{$region}{Option}{text_align},
		 text_size     => $Region{$region}{Option}{text_size},
		 text_align    => $Region{$region}{Option}{text_align},
		 reverse_color => $Region{$region}{Option}{reverse_color},
		 foreground    => $Region{$region}{Option}{foreground} || 'black',
		 background    => $Region{$region}{Option}{background} || 'white',
		};
      OPT: for my $opt (@opts) {
	if ( $opt =~ /^(\w+)=(.+)$/ ) {
	  my ($key, $value) = ($1, $2);
	  $key = $Config{Keys}{Opts}{$key} if exists $Config{Keys}{Opts}{$key};
	  $ref->{$key} = $value;
  	} else {
	  warn "Unparsable $opt";
	  next OPT;
	}
      }
      if (not defined $ref->{col} and $ref->{row}) {
	warn "No row and col?! '$_'";
	next LINE;
      }
      my ($col1, $col2, $row1, $row2);
      if ($ref->{col} =~ /^(\d+)..(\d+)$/) { $col1 = $1; $col2 = $2;      }
      else                                 { $col1 = $col2 = $ref->{col}; }
      if ($ref->{row} =~ /^(\d+)..(\d+)$/) { $row1 = $1; $row2 = $2;      }
      else                                 { $row1 = $row2 = $ref->{row}; }

      for my $col ($col1..$col2) {
	for my $row ($row1..$row2) {
	  my $new = {%{$ref}};
	  $new->{col} = $col;
	  $last_col = $col;
	  $new->{row} = $row;
	  $last_row = $row;
	  push @{ $Region{$region}{Content} }, $new;
	}
      }
      $last_row++;
    }
  }
  close DEF;
  Debug \%Region;
  return 1;
}

#-----------------------------------------------------------------------------
# Read the translations
sub ParseTextData ( $ ) {
  my $form = shift;
  my $file = "$Opt{DataDir}/$form";
  $form =~ s/\...$//;
  open TEXT, $file or die;
  while (<TEXT>) {
    chomp;
    /^\s*([\w]+:[\w]+)\s*=\s*(.*?)\s*$/ or next;
    my ($key, $value) = ($1, $2);
    $Text{"$form:$key"}=$value;
    Debug "Text{\"$form:$key\"}=$value";
  }
  close TEXT;
  return 1;
}

#=============================================================================
#=============================================================================
# Voodoo


#-----------------------------------------------------------------------------
# The magic recursive text lookup function. Add voodoo here.
sub Lookup ( $ ) {
  my $text = shift;
  Debug "Text:", $text, "Depth:", scalar @LookupStack;
  die "Lookup forms a loop, '", join ("'->'", @LookupStack) . "'\n"
    if grep /^\Q$text\E$/, @LookupStack;
  push @LookupStack, $text;

  if (defined $Text{"$Form:$text"}) {
    $text = $Text{"$Form:$text"};
  } elsif (defined $Text{$text}) {
    $text = $Text{$text};
  } elsif ($text =~ /(\#{(\w+:\w+)})/) {
    my ($t, $lookup) = ($1, "$Form:$2");
    if (exists $Text{$lookup}) {
      $text =~ s/$t/$Text{$lookup}/g;
    } else {
      warn "Missing lookup for '$t'\n";
      $Opt{Error} = 1;
    }
  } elsif ($text =~ /^(?:$Form:)?\w+:(\d+)$/) {
    $text = $1;
  } elsif ($text =~ /^$Form:\w+:([\$\@]\w+)$/) {
    $text = $1;
  } elsif ($text =~ /^\$(\w+)$|\$\{(\w+)\}/) {
    my $var = $1 || $2;
    my $key;
    # Variables all in CAPS are not per-form.
    if ($var eq uc $var)    { $key = $var;             }
    else                    { $key = "$Form.$ID:$var"; }
    Debug "Var:", $var;
    if (exists $CharData{$key}) {
      $text =~ s/\$\{?$var\}?/$CharData{$key}[0]/;
    } else {
      warn "Missing character data '$var'\n";
      $CharData{$key}[0] = "?";
      $Opt{Error} = 1;
      $text =~ s/\$\{?$var\}?/?/;
    }
  } elsif ($text =~ /^\@(\w+)$|\@\{(\w+)\}/) {
    my $var = $1 || $2;
    my $key = "$Form.$ID:$var";
    Debug "Var:", $var;
    if (exists $CharData{$key}) {
      my $value = shift @{ $CharData{$key} };
      $value = "" if not defined $value;
      $text =~ s/\@\{?$var\}?/$value/;
    } else {
      warn "Missing character data '$var'\n";
      $CharData{$key}[0] = "";
      #$Opt{Error} = 1;
      $text =~ s/\@$var//;
    }
  } elsif ( @LookupStack == 1 ) {
    warn "Missing translation for '$text'\n";
    $Text{$text} = "?";
    $Opt{Error} = 1;
    $text = "?";
  } else {
    Debug "End:", $text;
    pop @LookupStack;
    return $text;
  }
  Debug "After:", $text;
  $text = Lookup $text;
  pop @LookupStack;
  return $text;
}

#=============================================================================
#=============================================================================
# Embedding

#-----------------------------------------------------------------------------
# Embed a font.
sub Font ( $$ ) {
  my $p = shift;
  my $file = shift;
  $file = "$Opt{FontDir}/$file";
  return $Fonts{$file} if exists $Fonts{$file};
  my $font = $P->psfont("$file.pfb","$file.afm",'latin1');
  $Fonts{$file} = $font;
  return $font;
}

#=============================================================================
#=============================================================================
# The rendering functions.

#-----------------------------------------------------------------------------
# Top level function to print a character sheet.
sub Print () {
  my $file = "$File.pdf";
  Debug "Printing. File: $file";

  # create a new PostScript object
  $P = new PDF::API2::Lite;

  my $location = {
		  t => Margin,
		  l => Margin,
		  w => PageWidth-(2*Margin),
		  h => PageLength-(2*Margin),
		 };

  $CharData{PAGES}[0] = @Pages;
  $CharData{PAGE}[0]  = 0;
  $CharData{DATE}[0]  = localtime;
  $CharData{FILE}[0]  = $File;
  for my $form (@Pages) {
    my $page;
    ($Form, $ID, $page) = $form =~ /^([-\w]+)\.(\w+):(\w+)$/ or confess;
    Debug "Form:", $Form, "ID:", $ID, "Page:", $page;

    # Create a new page.
    $P->page($Opt{Paper});
    $CharData{PAGE}[0] ++;

    # Render the page in passes.
    for $Pass ("fill","line","text") {
      Debug "Pass:", $Pass;
      DrawRegion ("$Form:$page", $location);
    }

  } # End: for my $form (@Pages) {

  if ($Opt{Error}) {
    print "Errors found processing. No page produced.\n";
    exit 1;
  }

  $P->saveas($file);
  print "Output into $file\n";
}

#-----------------------------------------------------------------------------
# Render regions recursively.
sub DrawRegion ( $$ ) {
  my $region = shift;
  confess if not $region;
  #($Form) = $region =~ /^(.+)\./;

  my $loc = shift;

  Debug "Rendering", $region, "in", $loc, "Stack:", \@RenderStack;
  for (@RenderStack) {
    confess "Fatal, recursive definiton: @RenderStack"
      if $_ eq $region;
  }

  if (not exists $Region{$region}) {
    warn "Missing a region: $region\n";
    $Opt{Error} = 1;
    return;
  }

  push @RenderStack, $region;

  my $left   = $loc->{l};
  my $top    = $loc->{t};
  my $bottom = $loc->{t}+$loc->{h};
  my $right  = $loc->{l}+$loc->{w};
  my $width  = $loc->{w};
  my $height = $loc->{h};

  my $scale_w = $width  / $Region{$region}{Option}{col};
  my $scale_h = $height / $Region{$region}{Option}{row};

  ELEMENT: for my $ref ( @{ $Region{$region}{Content} }) {
    Debug "Content:", $ref->{content};
    if ($ref->{col} < 1) {
      warn "Column index for $region\->$ref->{content} is too low.\n";
      $Opt{Error} = 1;
      next ELEMENT;
    }
    if ($ref->{row} < 1) {
      warn "Row index for $region\->$ref->{content} is too low.\n";
      $Opt{Error} = 1;
      next ELEMENT;
    }
    if ( ($ref->{width} + $ref->{col}-1) > $Region{$region}{Option}{col}) {
      warn "Position and width of '$region\->$ref->{content}' is outside of bounding cell.\n";
      $Opt{Error} = 1;
      next ELEMENT;
    }
    if ( ($ref->{height} + $ref->{row}-1) > $Region{$region}{Option}{row}) {
      warn "Position and height of '$region\->$ref->{content}' is outside of bounding cell.\n";
      $Opt{Error} = 1;
      next ELEMENT;
    }
    my $l = $left + (($ref->{col}-1) * $scale_w);
    my $t = $top + (($ref->{row}-1) * $scale_h);
    my $w = $ref->{width} * $scale_w;
    my $h = $ref->{height} * $scale_h;
    my $loc = {l=>$l, t=>$t, w=>$w, h=>$h};
    if ($ref->{content} =~ /^region:(\w+)$/) {
      DrawRegion ("$Form:$1", $loc);
    } else {
      DrawFill ($loc, $ref);
      DrawBorders ($t, $l, $t+$h, $l+$w, $ref);
      DrawText "$region:$ref->{content}", $loc, $ref;
    }
  }

  DrawBorders ($top, $left, $bottom, $right, $Region{$region}{Option});

  pop @RenderStack;
}

#-----------------------------------------------------------------------------
# Draw the borders around a cell.
sub DrawBorders ( $$$$$ ) {
  return unless $Pass eq "line";
  my ($t, $l, $b, $r, $ref) = @_;
  StrokeColor("black");
  DrawBox  ($t,$l,$b,$r, $ref->{border})        if $ref->{border};
  DrawLine ($t,$l,$t,$r, $ref->{border_top})    if $ref->{border_top};
  DrawLine ($b,$l,$b,$r, $ref->{border_bottom}) if $ref->{border_bottom};
  DrawLine ($t,$l,$b,$l, $ref->{border_left})   if $ref->{border_left};
  DrawLine ($t,$r,$b,$r, $ref->{border_right})  if $ref->{border_right};
}

#-----------------------------------------------------------------------------
# Fill in a cell.
sub DrawFill ( $$ ) {
  return unless $Pass eq "fill";
  my ($where, $how) = @_;
  Debug "Where:", $where;
  Debug "How:", $how;
  my $color = $how->{reverse_color} ?
    $how->{foreground} : $how->{background};
  Debug "Fill color:", $color;

  FillColor($color);
  $P->rectxy(
	     Point($where->{l}, $where->{t}),
	     Point($where->{l}+$where->{w}, $where->{t}+$where->{h})
	    );
  $P->fill;
}

#-----------------------------------------------------------------------------
# Draw the text in a cell.
sub DrawText ( $$$ ) {
  return unless $Pass eq "text";
  my ($text, $where, $how) = @_;
  Debug "Text:", $text;
  Debug "Where:", $where;
  Debug "How:", $how;

  my $align = $how->{text_align} || $Opt{FontAlign};

  my $color= $how->{reverse_color} ?
    $how->{background} : $how->{foreground};
  Debug "Text color:", $color;

  my $font = $how->{text_font} || $Opt{FontType};
  my $size = $how->{text_size} || $Opt{FontSize};

  $text = Lookup $text;

  $font = Font ($P, $font);
  FillColor($color);

  if ($align =~ /^c/) {
    # Center.
    $where->{l} += ($where->{w})/2;
    $align = 1;
  } elsif ($align =~ /^l/) {
    # Left.
    $where->{l} += $Opt{Padding};
    $align=0;
  } else {
    # Right.
    $where->{l} += $where->{w} - $Opt{Padding};
    $align=2;
  }


  my $left = $where->{l};
  while ($text =~ s/^\.//g) {
    $left += $Opt{Padding};
  }

  $P->print ($font,
	     $size,
	     Point($left, $where->{t}+$where->{h}-$Opt{Baseline}),
	     0,
	     $align,
	     $text);
  return 1;
}

#-----------------------------------------------------------------------------
# Draw a box someplace.
sub DrawBox ( $$$$$ ) {
  my ($top,$left,$bottom,$right,$code) = @_;

  for my $width (split //, $code) {
    $width *= $Opt{LineWidth};
    $P->linewidth( $width );
    $P->rectxy(Point($left,$top), Point($right,$bottom));
    $P->stroke;
    $top    -= $width*2;
    $left   -= $width*2;
    $bottom += $width*2;
    $right  += $width*2;
  }
}

#-----------------------------------------------------------------------------
# Draw a line someplace.
sub DrawLine ( $$$$$ ) {
  my ($top,$left,$bottom,$right,$code) = @_;

  for my $width (split //, $code) {
    $width *= $Opt{LineWidth};
    $P->linewidth( $width );
    $P->move (Point($left,$top));
    $P->line (Point($right,$bottom));
    $P->stroke;
    $top    -= $width*2;
    $left   -= $width*2;
    $bottom += $width*2;
    $right  += $width*2;
  }
}
#-----------------------------------------------------------------------------
# Set the fill color.
sub FillColor ( $ ) {
  my $color = shift;
  Debug "Color:", $color;
  $P->fillcolor($color);
}
#-----------------------------------------------------------------------------
# Set the stroke color.
sub StrokeColor ( $ ) {
  my $color = shift;
  Debug "Color:", $color;
  $P->strokecolor($color);
}

#=============================================================================
#=============================================================================
# Print the data structures.

#-----------------------------------------------------------------------------
# Print the translations.
sub PrintFormTranslation ( ) {
  for my $region (sort keys %Region) {
    for my $ref (sort {$a->{content} cmp $b->{content}}
		      @{ $Region{$region}{Content} }) {
      next unless $ref->{content} =~ /^region:/;
      print "$region:$ref->{content}=",
	     Lookup ("$region:$ref->{content}"),
	     "\n";
    }
    print "\n";
  }
}

#-----------------------------------------------------------------------------
# Print the character data.
sub PrintCharacterData ( ) {
  print "START CHARCTER DATA\n";
  for my $key (sort keys %CharData) {
    print "$key = $CharData{$key}\n";
  }
  print "END\n";
}


#=============================================================================
#=============================================================================
# Debug functions.

#-----------------------------------------------------------------------------
# Are we in debug mode?
sub _debugging () {
  return 1 if $main::Opt{Debug};
  return 0;
}

#-----------------------------------------------------------------------------
# Print out debug values.
sub Debug ( @ ) {
  return 1 if not _debugging;

  my $line    = (caller(0))[2];
  my $sub     = (caller(1))[3] || "main::main";
  $sub = "main::$sub" if $sub !~ /::/;

  my @dump;
  my @arg = @_;
  for (@arg) {
    if (ref $_) {
      push @dump, Dumper $_;
      $_ = "<DUMPED>";
    } elsif (not defined $_) {
      $_ = '<undef>';
    }
  }

  printf STDERR "D:%3d:%-15s %s\n", $line, $sub, join " ", @arg;
  for my $dump (@dump) {
    for my $dl (split /\n/, $dump) {
      printf STDERR "D:%3d:%-15s %s\n", $line, $sub, $dl;
    }
  }

}
